@@ -10,8 +10,8 @@ import warnings
 import numpy as np
 from tlz import concat, frequencies
 
-from dask.array.core import Array
 from dask.array.numpy_compat import AxisError
+from dask.base import is_dask_collection, tokenize
 from dask.highlevelgraph import HighLevelGraph
 from dask.utils import has_keyword, is_arraylike, is_cupy_type, typename
 
@@ -42,7 +42,7 @@ def meta_from_array(x, ndim=None, dtype=None):
     """
     # If using x._meta, x must be a Dask Array, some libraries (e.g. zarr)
     # implement a _meta attribute that are incompatible with Dask Array._meta
-    if hasattr(x, "_meta") and isinstance(x, Array):
+    if hasattr(x, "_meta") and is_dask_collection(x) and is_arraylike(x):
         x = x._meta
 
     if dtype is None and x is None:
@@ -61,11 +61,11 @@ def meta_from_array(x, ndim=None, dtype=None):
 
     if isinstance(x, list) or isinstance(x, tuple):
         ndims = [
-            0
-            if isinstance(a, numbers.Number)
-            else a.ndim
-            if hasattr(a, "ndim")
-            else len(a)
+            (
+                0
+                if isinstance(a, numbers.Number)
+                else a.ndim if hasattr(a, "ndim") else len(a)
+            )
             for a in x
         ]
         a = [a if nd == 0 else meta_from_array(a, nd) for a, nd in zip(x, ndims)]
@@ -210,7 +210,20 @@ def _check_dsk(dsk):
     assert all(isinstance(k, (tuple, str)) for k in dsk.layers)
     freqs = frequencies(concat(dsk.layers.values()))
     non_one = {k: v for k, v in freqs.items() if v != 1}
-    assert not non_one, non_one
+    key_collisions = set()
+    # Allow redundant keys if the values are equivalent
+    collisions = set()
+    for k in non_one.keys():
+        for layer in dsk.layers.values():
+            try:
+                key_collisions.add(tokenize(layer[k]))
+            except KeyError:
+                pass
+        if len(key_collisions) >= 2:
+            collisions.add(k)
+        key_collisions.clear()
+
+    assert len(collisions) == 0, {k: non_one[k] for k in collisions}
 
 
 def assert_eq_shape(a, b, check_ndim=True, check_nan=True):
@@ -255,7 +268,7 @@ def _get_dt_meta_computed(
     x_meta = None
     x_computed = None
 
-    if isinstance(x, Array):
+    if is_dask_collection(x) and is_arraylike(x):
         assert x.dtype is not None
         adt = x.dtype
         if check_graph:
@@ -426,6 +439,8 @@ def arange_safe(*args, like, **kwargs):
 
 
 def _array_like_safe(np_func, da_func, a, like, **kwargs):
+    from dask.array.core import Array
+
     if like is a and hasattr(a, "__array_function__"):
         return a
 
